import base64
import csv
import functools
import json
import logging
import sys
import warnings
from collections import defaultdict
from copy import deepcopy
from datetime import datetime
from functools import lru_cache, partial
from io import StringIO
from itertools import groupby
from json import JSONDecodeError
from pathlib import Path
from string import Formatter
from typing import Optional

from argus_api.api.cases.v2.case import (
    advanced_case_search,
    create_case,
    add_comment,
    close_case,
    add_attachment,
    update_case,
    request_workflow as request_case_workflow,
    acknowledge_workflow as acknowledge_case_workflow,
    list_workflows as list_case_workflows,
)
from argus_api.api.customers.v2.customer import get_customer
from argus_api.api.events.v1.aggregated import update_events
from jinja2 import Environment, FileSystemLoader, select_autoescape

from argus_cli import run
from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_cli.utils.time import datetime_to_week_number
from argus_plugins import argus_cli_module
from argus_plugins.cases.autocreate_case_output_template import (
    AUTOCREATE_OUTPUT_TEMPLATE,
)
from argus_plugins.cases.utils import STATUSES, PRIORITIES, CASE_TYPES, format_tags_list

# Cache the function to avoid multiple calls to the backend
from argus_plugins.datastore.synchronize import set_data_with_buffer
from argus_plugins.enrichments import EVENTS_ENRICHMENTS, enrich_data

get_customer = lru_cache(maxsize=None)(get_customer)

#: Status assigned when reopening cases due to new events
REOPENED_CASE_STATUS = "pendingCustomer"

#: Action to take when new events are added to a closed case
#: - "reopen": the case is reopened.
#:   This is the default unless the autoreport is set to close cases on creation.
#: - "silent": events are added to the case, which remains closed.
#:   This is the default when the autoreport is set to close cases on creation.
#: - "new-case": create a new case with the new events, bypassing any timeout set.
CLOSED_CASE_UPDATE_METHODS = ("reopen", "silent", "new-case")


def _case_watcher_json(
    event: dict = None,
    case_watcher_from_field: str = None,
    case_watcher: str = None,
    watcher_type: str = "email",
    verbose: bool = False,
):
    """Creates an argus supported structure for case watchers"""
    watchers = [case_watcher] if case_watcher is not None else []
    if case_watcher_from_field is not None:
        watcher = value_from_nested_key(case_watcher_from_field, event)
        # Using extend avoids the possibility of adding None
        watchers.extend([watcher] if watcher is not None else [])

    structured_watchers = []
    for watcher in watchers:
        structured_watchers.append(
            {
                "type": watcher_type,
                "destination": watcher,
                "verbose": verbose,
            }
        )
    return structured_watchers


def customer_language(customer: dict) -> str:
    """Convenience function to get a customer's language"""
    try:
        return customer["language"]
    except KeyError:
        log.debug("Fetching language from Argus for {shortName}".format_map(customer))
        return get_customer(idOrShortName=customer["shortName"])["data"][
            "language"
        ].upper()


def value_from_nested_key(key: str, d: dict):
    """Allows the user to reference a variable with normal python string-format syntax.

    Example:
        >>> value_from_nested_key("a[b][c]", {"a": {"b": {"c": 1337}}})
        1337
        >>> value_from_nested_key("a", {"a": 42})
        42
    """
    try:
        return Formatter().get_field(key, (), d)[0]
    except KeyError:
        return None
    except AttributeError:
        return None


def _get_field(field: str, event: dict):
    """Returns a hashable key for us to use

    Dicts and lists are not hashable, so we have to make them hashable.
    This is a representation problem that could have been fixed by having objects
    representing the data instead of dicts.
    """
    new = value_from_nested_key(field, event)

    if isinstance(new, dict):
        # Dicts are not hashable
        new = tuple(sorted(new.items()))
    elif isinstance(new, list):
        # List ain't hashable either.
        new = tuple(new)
    elif new is None:
        log.warning(f"field {field} missing from event #{event.get('id')}")

    return new


def group_inputs(data: list, group_by: list) -> dict:
    """Groups the input based on the group_by parameter.

    The first field in group_by will be the outermost group in the dict.

    :param data: The input
    :param group_by: Field to group by
    :returns: A nested group dict
    """

    def _key(event):
        return tuple(_get_field(field, event) for field in group_by)

    # Has to be sorted, see groupby()'s documentation.
    sorted_data = sort_input(data, group_by)

    return {
        # groupby() returns the group as an iterator, so we have to unpack it.
        group_key: list(group)
        for group_key, group in groupby(sorted_data, key=_key)
    }


def groupby_jinja_filter(value, key):
    """Custom groupby for jinja templates that supports dots in the keys

    This is implemented as pallets have decided not to support this syntax in
    jinja. See https://github.com/pallets/jinja/issues/1142

    This implementation wraps around another input grouping that already exists
    in this codebase, group_inputs(), and modifies the output from that to match
    Jinja's own implementation of groupby. For Jinja's implementation, see:
    https://github.com/pallets/jinja/blob/master/src/jinja2/filters.py#L930

    :param value: The iterable with the data that should be grouped
    :param key: A python-style dictionary reference. See value_from_nested_key()
                for examples.
    """

    # A note on the implementation seen under:
    #       group_inputs() allows for multiple keys, which is not supported
    #       by jinja's groupby. Thus we have wrap the key in a list and unpack
    #       the resulting group key that was returned.
    return [
        (group_key[0], group) for group_key, group in group_inputs(value, [key]).items()
    ]


def sort_input(data: list, sort_by: list) -> list:
    """Sorts the given inputs based on the given list.

    The first field in sort_by will be the one with most precedence (primary key).

    :param data: The input
    :param sort_by: Fields to sort by
    :returns: A sorted list of events
    """

    def _key(event):
        # "None" is not sortable, so adding an extra sorting value for that.
        k = []
        for field in sort_by:
            val = _get_field(field, event)
            k.append((val is None, val))
        return k

    return sorted(data, key=_key)


def group_data(data: list, groups: list):
    """Gives the data groups based on the customer and groups"""
    customer_events = group_inputs(data, ["customerInfo"])
    for customer, data in customer_events.items():
        # Workaround because group_inputs()'s key is a tuple.
        customer = dict(customer[0])
        log.info('Handling data for customer "{}"'.format(customer["shortName"]))

        groupings = group_inputs(data, groups)
        for group, data in groupings.items():
            # Argus requires the group to be a string
            yield (customer, str(group), data)


def put_data_in_case(
    data: list,
    customer: dict,
    group: str,
    find_existing_case: callable,
    create_description: callable,
    create_comment: callable,
    create_case: callable,
    event_associator: callable,
    on_case_created: callable,
    on_case_updated: callable,
    use_fields: bool,
    set_initial_comment: bool = False,
    new_case_on_closed: bool = False,
    test_data: bool = False,
    silent_update: bool = False,
):
    """Puts the data into the given group for the given customer.

    :param data: List of data
    :param customer: The customer object to associate with the data
    :param group: The group that the data belongs to
    :param find_existing_case:
    :param create_description:
    :param create_comment:
    :param create_case:
    :param event_associator:
    :param set_initial_comment: Add internal comment on case creation
    :param on_case_created: callback used whenever a new case is created
    :param on_case_updated: callback used whenever an existing case is updated
    :param new_case_on_closed: if set to True, create a new case when there is an existing
      case that is in a closed state. Used to implement the "new-case" method for the
      "--closed-case-update" option.
    :param test_data: whether to include test cases when searching for existing cases
    :param silent_update: if set, do not create a comment upon adding new events to an
      existing case.
    :returns: The case that got commented or created.
    """
    # If new_case_on_closed is true, then we do not want to find an existing closed case.
    case = find_existing_case(
        customer,
        group,
        include_closed_cases=not new_case_on_closed,
        include_test_cases=test_data,
        use_fields=use_fields,
    )

    if case and case["status"] == "closed" and new_case_on_closed:
        log.info(f"Ignoring existing closed case #{case['id']}")
        case = None
    description = create_description(data, customer, comment=True if case else False)
    log.debug("Description for the given group: {}".format(description))
    if case:
        # If ``silent_update`` is provided, set the ``comment`` variable to ``None`` to exclude it from the request body
        # created in the ``update_case`` call
        comment = None if silent_update else description
        on_case_updated(case, comment)
    else:
        case = create_case(customer, group, description, data[0])
        log.info("Created new case #{id}".format_map(case))
        if set_initial_comment:
            initial_comment = create_description(
                data, customer, comment=False, internal=True
            )
            create_comment(case["id"], comment=initial_comment, internal=True)
            log.debug(
                "Added internal comment after creating case #{id}".format_map(case)
            )
        on_case_created(case)
        log.debug("Executed on_case_created callback for case #{id}".format_map(case))

    event_associator(case["id"], data)
    log.debug(
        "Associated these events with case {id}:\n{events}".format(
            id=case["id"], events=[event["id"] for event in data]
        )
    )

    return case


def print_results(events: list, cases: dict, dry: bool):
    if dry:
        print("\nThis was a dry run, script would have done the following:\n")
    stats = ""
    for customer, group, case, _ in cases.values():
        stats += "\n\t{} (#{}) - {} - {}".format(
            case["subject"], case["id"], customer, group
        )

    print(
        "Analyzed {event_amount} events, and associated them with "
        "{case_amount} cases.\n\t{stats}".format(
            event_amount=len(events), case_amount=len(cases), stats=stats
        )
    )

    for customer, group, case, events in cases.values():
        subject = f"{case['subject']} (#{case['id']})"
        case_output = AUTOCREATE_OUTPUT_TEMPLATE.substitute(
            subject=subject,
            subject_line="=" * len(subject),
            description=case["description"],
            description_line="-" * len("Description"),
            associated_events_line="-" * len("Associated events"),
            associated_events="\n".join(event["id"] for event in events),
        )
        print(case_output)


def _existing_case(
    customer: dict,
    group: str,
    timeout,
    key,
    use_fields: bool,
    include_closed_cases=True,
    include_test_cases=False,
    case_search_callable=None,
):
    """Returns an existing case based on the customer, key and group"""
    # This needs to be resolved at runtime as the tests patch this function, which would not be properly applied
    # if specified directly in the default parameter.
    case_search_callable = case_search_callable or advanced_case_search
    # If it turns out that there are a lot of search requests, then
    # this could probably be changed in favor of a single request to the "key"
    # and then filter for customer and group internally.

    tag_or_field_kwargs = {}
    subcriteria = []

    if use_fields:
        tag_or_field_kwargs["field"] = [
            {"field": "case-autocreate-key", "values": [key]}
        ]
        subcriteria.append(
            {
                "required": True,
                "field": [{"field": "case-autocreate-group", "values": [group]}],
            }
        )
    else:
        tag_or_field_kwargs["tag"] = [{"key": "case-autocreate-key", "values": [key]}]
        subcriteria.append(
            {
                "required": True,
                "tag": [{"key": "case-autocreate-group", "values": [group]}],
            }
        )

    if not include_closed_cases:
        subcriteria.append({"exclude": True, "status": "closed"})

    cases = case_search_callable(
        startTimestamp=int(
            timeout.timestamp() * 1000
        ),  # Only get cases within the timeout
        timeFieldStrategy=["createdTimestamp"],
        customerID=[customer["id"]],
        includeTestData=include_test_cases,
        **tag_or_field_kwargs,
        # Tags are OR-ed together, we want to AND them
        subCriteria=subcriteria,
    )["data"]

    # There should never be more than one case returned by this. If there is, then just pick the first one, as it is not practically possible to
    # precisely determine which should be used. If order is important, then the criteria should be updated to only return one case.
    return cases[0] if cases else None


@register_command(extending="cases", module=argus_cli_module)
def autocreate(
    data: sys.stdin,
    key: str,
    template_folder: Path,
    group_by: list = [],
    sort_by: list = [],
    timeout: datetime = None,
    case_title: str = None,
    case_title_no: str = None,
    case_title_en: str = None,
    case_title_format: ["fmtstr", "jinja"] = "fmtstr",
    case_status: STATUSES = "pendingCustomer",
    case_priority: PRIORITIES = "medium",
    case_type: CASE_TYPES = "securityIncident",
    case_service: str = "ids",
    case_category: str = None,
    case_watcher: str = None,
    case_watcher_verbose: bool = False,
    case_watcher_from_field: str = None,
    skip_notifications: bool = False,
    initial_internal_comment: bool = False,
    internal_case: bool = False,
    explicit_access: list = [],
    request_soc_analysis: bool = False,
    request_workflow: str = None,
    acknowledge_workflow: str = None,
    workflow_comment: str = None,
    dry: bool = False,
    close_after_create: bool = False,
    enrich: EVENTS_ENRICHMENTS = None,
    tags: dict = None,
    fields: dict = None,
    attach_events: ["json", "csv"] = False,
    status_on_update: STATUSES = None,
    closed_case_update: CLOSED_CASE_UPDATE_METHODS = None,
    silent_update: bool = False,
    test_data: bool = False,
    send_to_qa: bool = False,
    use_fields: bool = False,
):
    """A tool for automatically creating a case based on events and similar.

    Customers will automatically be extracted from the given data.
    There will be one case per group, meaning that if you group by email addresses,
    you'll get one case per email address.

    Examples:
        New case every time the script runs:
        $ argus-cli ./data.json "protocol-data" ./
            --group-by protocol
            --case-title "We've spotted some protocols"

        New case every week based on the alarm ID
        $ echo "<List of JSON objects>"| argus-cli cases autocreate "scary-events" ./templates/
            --group-by attackInfo[alarmId]
            --timeout "1 week"
            --case-title "This week, scary stuff happened."

    :param data: JSON-data to parse. Will typically be passed via stdin.
    :param key: A unique ID for this autocreate instance. Will be associated to the case as a tag.
    :param template_folder: Folder with JINJA templates. Filenames are: <key>.<language>.html
    :param group_by: Identifiers in the data to group by (Unique cases will be created per group).
    :param sort_by: Identifiers in the data to sort by
    :param timeout: The timeframe between new cases. If not specified a new case will be created every run.
    :param case_title: The title of the created case, ignoring customer language settings.
        Can be used with python string formatting. Default: "Autocreated based on group {group}". Incompatible
        with --case-title-[en|no].
    :param case_title_no: The title of the created case for customers preferring the Norwegian language.
        Incompatible with --case-title. If set, --case-title-en must also be set.
    :param case_title_en: The title of the created case for customers preferring the English language.
        Incompatible with --case-title. If set, --case-title-no must also be set.
    :param case_title_format: if set to "jinja", title options will be processed as jinja templates.
    :param case_status: The status of the created case.
    :param case_priority: The priority of the created case.
    :param case_type: The type of the created case.
    :param case_service: The service of the created case.
    :param case_category: The category of the created case.
    :param case_watcher: Watcher that should be added to the case.
    :param case_watcher_from_field: String that points to where e-mail address
        for case watcher is in the event. The format is like python's dict
        access. For example: properties[property.name]
    :param skip_notifications: If set, no notifications will be sent on case creation.
    :param initial_internal_comment: Creates a new internal comment on case creation. This can be added with a block checking for "internal" in your template.
    :param internal_case: Makes new cases only visible internally for techs.
    :param explicit_access: If set, the case will be restricted to the provided user(s) (with write access level). Can be repeated.
    :param request_soc_analysis: If set, immediately request the "socAnalysis" workflow. Note: If the "socAnalysis" workflow is already requested, this request will be ignored.
    :param request_workflow: Request a new workflow of the specified type. Note: If workflow is already requested, this request will be ignored.
    :param acknowledge_workflow: Acknowledge a new workflow of the specified type. Note: If this workflow is not active, this request will be ignored.
    :param workflow_comment: An optional comment to pass when requesting or acknowledging a workflow
    :param dry: If set, no data will be committed.
    :param close_after_create: If set, the case will be closed immediately after creation.
    :param list enrich: optional enrichment to perform on the events.
    :param tags: key-value tags to add to the created case. Expects JSON key-values {"tag": "value"}, for mutliple tags use a JSON object or list : {"tag1" : "tag1-value", "tag2": "tag2-value"}
    :param fields: Case fields to set con created cases. Note! Field must already exist in Argus beforehand to be able to be set. Expecs JSON in the format `{"tag": "value"}`.
    :param attach_events: Format to attach events in. If left unset, no events
            are attached.
    :param status_on_update: If set, case status to set when events are added to an existing case.
    :param closed_case_update: Action to take when new events are added to a closed case. "reopen": reopen the case. "silent": leave the case closed. "new-case": create a new case.
    :param silent_update: if set, do not add a comment upon adding new events to an existing case.
    :param test_data: create the case as test data. Note: If a non-test case with matching --group-by value already
        exists, new events will be associated to that existing case even though --test-data is passed here. To verify
        that a test case will be used, perform a dry run first.
    :param send_to_qa: Create the case as normal, but set it as unpublished. This will appear in the QA process.
    :param use_fields: If set, use case fields to identify existing cases instead of tags. NOTE!: This option will be removed when tag handling is removed from this command as a whole.
    """
    # Encapsulate data into functions to avoid tramp data
    timeout = timeout or datetime.now()
    if internal_case:
        warnings.warn(
            "--internal-case is deprecated. If you intend for the case to be reviewed by the SOC before publishing, use --send-to-qa."
        )
    if internal_case and send_to_qa:
        raise ValueError("Only one of --internal-case and --send-to-qa can be present.")
    _add_comment = partial(
        add_comment,
        notification={"skipEmail": skip_notifications, "skipSMS": skip_notifications},
    )

    default_case_title = "Autocreated based on group {group}"

    jinja_env = Environment(
        # jinja doesn't accept Path objects
        loader=FileSystemLoader(str(template_folder.expanduser().absolute())),
        autoescape=select_autoescape(["html", "xml"]),
    )
    jinja_env.filters["custom_groupby"] = groupby_jinja_filter
    jinja_env.filters["week_num"] = datetime_to_week_number

    # handle title options
    if case_title:
        if case_title_en or case_title_no:
            sys.exit(
                "please use EITHER --case-title OR --case-title-en and --case-title-no"
            )
        case_title_loc = {"ENGLISH": case_title, "NORWEGIAN": case_title}
    elif case_title_en or case_title_no:
        if not (case_title_en and case_title_no):
            sys.exit("please specify BOTH --case-title-en AND --case-title-no")
        case_title_loc = {"ENGLISH": case_title_en, "NORWEGIAN": case_title_no}
    else:
        case_title_loc = {
            "ENGLISH": default_case_title,
            "NORWEGIAN": default_case_title,
        }
    if case_title_format == "jinja":
        default_case_title_jinja = "Autocreated based on group {{group}}"
        for lang, title in case_title_loc.items():
            if title == default_case_title:
                case_title_loc[lang] = default_case_title_jinja
            case_title_loc[lang] = jinja_env.from_string(case_title_loc[lang])

    def _create_description(
        events: list, customer: dict, comment: bool = False, internal: bool = False
    ) -> str:
        """Creates a case description from a template file"""
        # Argus returns long names for languages, let's use 2 letter versions.
        language_map = {"NORWEGIAN": "no", "ENGLISH": "en"}

        template = jinja_env.get_template(
            "{key}.{language}.html".format(
                key=key,
                language=language_map[customer_language(customer)],
                current_time=datetime.now(),
                timeout=timeout,
            )
        )
        return template.render(data=events, comment=comment, internal=internal)

    def _create_subject(data: dict, group: str) -> str:
        """creates the case's subject based on title options."""
        if case_title_format == "fmtstr":
            # use a defaultdict for properties to avoid KeyErrors
            data_defaults = deepcopy(data)
            data_defaults["properties"] = defaultdict(str, data_defaults["properties"])
            return case_title_loc[customer_language(customer)].format(
                group=group, **data_defaults
            )
        elif case_title_format == "jinja":
            return case_title_loc[customer_language(customer)].render(
                group=group, current_time=datetime.now(), timeout=timeout, **data
            )
        else:
            raise ValueError(
                f"unexpected --case-title-format value : {case_title_format}"
            )

    def _create_case(customer: dict, group: str, description: str, data: dict):
        """Creates a new case and adds tags to it.

        :param customer: The customer to create the case for.
        :param group: The group that the case is created from. Used for tagging and the title.
        :param description: Description of the case
        :param data: The data of the event. Used for the title.
        """
        case_data = dict(
            customerID=customer["id"],
            service=case_service,
            category=case_category,
            type=case_type,
            priority=case_priority,
            status=case_status,
            notification={
                "skipEmail": skip_notifications,
                "skipSMS": skip_notifications,
            },
            publish=(not internal_case),
            watchers=_case_watcher_json(
                event=data,
                case_watcher_from_field=case_watcher_from_field,
                case_watcher=case_watcher,
                verbose=case_watcher_verbose,
            ),
            subject=_create_subject(data, group=group),
            description=description,
            testData=test_data,
        )
        if internal_case:
            case_data["accessMode"] = "readRestricted"
            case_data["status"] = "pendingSoc"
        elif send_to_qa:
            case_data["status"] = "pendingSoc"
            case_data["publish"] = False
        elif explicit_access:
            case_data["accessMode"] = "explicit"
            case_data["aclMembers"] = [
                {"userOrGroup": u, "level": "write"} for u in explicit_access
            ]

        case_data["fields"] = [
            {"field": "case-autocreate-key", "value": key},
            {"field": "case-autocreate-group", "value": group},
            *fields,
        ]

        created_by_cron_tag = next(
            (t for t in tags if t["key"] == "CreatedByCron"), None
        )
        if created_by_cron_tag:
            case_data["fields"].append(
                {"field": "created-by-cron", "value": created_by_cron_tag["value"]}
            )

        case_data["tags"] = [
            {"key": "case-autocreate-key", "value": key},
            {"key": "case-autocreate-group", "value": group},
            *tags,
        ]

        case = create_case(**case_data)["data"]

        return case

    def _event_associator(case_id: int, events: list):
        """Associates the events to the case"""
        buffered_update = set_data_with_buffer(
            update_events, buffered_argument="eventIdentifiers"
        )
        buffered_update(
            eventIdentifiers=[event["id"] for event in events],
            update={
                "comment": "Automatically assessed by cases autocreate "
                "(Key: {})".format(key),
                "associateToCase": case_id,
            },
        )

        if attach_events:
            # In some cases, the customer wants to get the raw data of the event.
            if attach_events == "json":
                data = StringIO(json.dumps(events))
            elif attach_events == "csv":
                modified_events = [
                    {
                        "Event ID": event["id"],
                        "Customer": event["customerInfo"]["name"],
                        "Start time": event["startTimestamp"],
                        "End time": event["endTimestamp"],
                        "Source IP": event["source"]["networkAddress"]["address"],
                        "Source port": event["source"]["port"],
                        "Destination IP": event["destination"]["networkAddress"][
                            "address"
                        ],
                        "Destination port": event["destination"]["port"],
                        "Signature": event["attackInfo"]["signature"],
                        "Alarm ID": event["attackInfo"]["alarmID"],
                        "Description": event["attackInfo"]["alarmDescription"],
                        "Location": event["location"]["name"],
                        "Severity": event["severity"],
                        "Count": event["count"],
                    }
                    for event in events
                ]

                data = StringIO()
                csv_writer = csv.DictWriter(data, modified_events[0].keys())
                csv_writer.writeheader()
                csv_writer.writerows(modified_events)

            add_attachment(
                caseID=case_id,
                name=f"events.{attach_events}",
                mimeType=f"text/{attach_events}",
                data=base64.b64encode(data.getvalue().encode()).decode(),
            )

    def _on_case_created(case: dict):
        """Called whenever a new case is created"""
        if close_after_create:
            id_ = case["id"]
            if not dry:
                close_case(
                    caseID=id_,
                    notification={
                        "skipEmail": True,
                        "skipSMS": True,
                    },
                )
            log.info(f"Closed case #{id_}")
        if request_soc_analysis and request_workflow != "socAnalysis":
            _request_workflow(case, "socAnalysis")
        if request_workflow:
            _request_workflow(case, request_workflow, workflow_comment)
        elif acknowledge_workflow:
            _acknowledge_workflow(case, acknowledge_workflow, workflow_comment)

    def _on_case_updated(case: dict, comment: Optional[str]):
        """Called when a case is updated

        (when a case already existed, as opposed to _on_case_created)

        :param case: Updated case
        :param comment: Comment to add to case along with update
        """
        id_ = case["id"]
        if case["status"] == "closed" and closed_case_update == "reopen":
            reopen_status = status_on_update or REOPENED_CASE_STATUS
            if not dry:
                update_case(id_, status=reopen_status, comment=comment)

            info_log_msg = f"reopened case #{id_} (set to {reopen_status})"
            if comment is not None:
                info_log_msg += f" and added comment {comment}"

            log.info(info_log_msg)
        elif status_on_update and (case["status"] != status_on_update):
            if not dry:
                update_case(id_, status=status_on_update, comment=comment)

            info_log_msg = f"set case #{id_} to {status_on_update}"
            if comment is not None:
                info_log_msg += f" with comment {comment}"

            log.info(info_log_msg)

        elif comment:
            if not dry:
                _add_comment(caseID=id_, comment=comment)

            log.info(f"Added comment to case #{id_}: {comment}")

        if request_workflow:
            _request_workflow(case, request_workflow, workflow_comment)
        elif acknowledge_workflow:
            _acknowledge_workflow(case, acknowledge_workflow, workflow_comment)

    def _request_workflow(case: dict, flow: str, comment: str = None):
        """Request a workflow"""
        id_ = case["id"]
        if not dry:
            for w in list_case_workflows(caseID=id_)["data"]:
                if w["workflow"] == flow and w["state"] != "acknowledged":
                    log.info(
                        f"Ignoring request of workflow {flow} which is already active"
                    )
                    return
            request_case_workflow(id_, flow, comment)
        log.info(f"Requested workflow {flow} for case #{id_}")

    def _acknowledge_workflow(case: dict, flow: str, comment: str = None):
        """Request a workflow"""
        id_ = case["id"]
        target = None
        if not dry:
            for w in list_case_workflows(caseID=id_)["data"]:
                if w["workflow"] == flow and not w["acknowledgedByUser"]:
                    target = w
            if target:
                acknowledge_case_workflow(id_, flow, comment)
                log.info(f"Acknowledged workflow {flow} for case #{id_}")
            else:
                log.info(f"Ignoring acknowledge of workflow {flow} which is not active")
        else:
            log.info(f"Acknowledged workflow {flow} for case #{id_}")

    if not closed_case_update:
        closed_case_update = "silent" if close_after_create else "reopen"
    if dry:
        log.info("--- Running in DRY mode! No data will be committed. ---")
        log.setLevel(logging.DEBUG)
        for handler in log.handlers:
            handler.setLevel(logging.DEBUG)
        log.debug("Due to dry mode, debug output has been enabled.")
    if test_data:
        log.info(
            "--- Running in TEST-DATA mode! Cases will be created as test data. ---"
        )

    # process the tags early to fail before the case has been created
    try:
        if tags:
            tags = format_tags_list(tags)
        else:
            tags = []
    except ValueError:
        log.error(f"Invalid tags format: {tags}")
        exit(1)

    try:
        if fields:
            fields = format_tags_list(fields, key_str="field")
        else:
            fields = []
    except ValueError:
        log.error(f"Invalid fields format: {fields}")
        exit(1)

    try:
        data = json.load(data)  # Input is a file, so we'll have to parse it.
    except JSONDecodeError as e:
        if e.pos == 0:
            raise ValueError(
                "The supplied JSON input is empty. "
                "Your event-search most likely returned no result."
            ) from e
        else:
            raise e

    log.info("Received {} events".format(len(data)))

    data = sort_input(data, sort_by)

    if enrich is not None:
        log.debug(f"Enriching events with {' '.join(enrich)}")
        enrich_data(data, enrich)

    cases = {}

    putter = partial(
        put_data_in_case,
        find_existing_case=functools.partial(_existing_case, timeout=timeout, key=key),
        create_description=_create_description,
        create_comment=_add_comment if not dry else lambda *a, **kw: None,
        create_case=(
            _create_case
            if not dry
            else lambda *a, **kw: {"id": f"DRY {len(cases.values())}"}
        ),
        event_associator=_event_associator if not dry else lambda *a, **kw: None,
        on_case_created=_on_case_created,
        on_case_updated=_on_case_updated,
        set_initial_comment=initial_internal_comment,
        new_case_on_closed=(closed_case_update == "new-case"),
        test_data=test_data,
        silent_update=silent_update,
        use_fields=use_fields,
    )

    for customer, group, event in group_data(data, group_by):

        # We should already have events grouped by customer here, so this is
        # just an extra verification
        if any(customer["shortName"] != e["customerInfo"]["shortName"] for e in event):
            raise RuntimeError(
                "Not all events belong to the specified "
                + f"customer {customer['shortName']}: {event}"
            )

        case = putter(data=event, customer=customer, group=group)
        if dry:
            # [0] here to replicate behaviour in put_data_in_case.
            case["subject"] = _create_subject(event[0], group=group)
            case["description"] = _create_description(event, customer, group)

        cases[case["id"]] = (customer["shortName"], group, case, event)

    print_results(data, cases, dry)


if __name__ == "__main__":
    run(autocreate)



from typing import Callable
import warnings
from argus_api.pagination import (
    LimitOffsetPaginator,
)  # exposed here for backward compatibility
from argus_api.pagination import offset_paginated as _offset_paginated


def offset_paginated(func: Callable) -> Callable:
    warnings.warn(
        "offset_paginated has been moved to the argus_api package and will be removed from argus_cli.",
        category=DeprecationWarning,
    )
    return _offset_paginated(func)



from datetime import datetime, timedelta

from argus_api.api.cases.v2.case import advanced_case_search
from dateutil.relativedelta import relativedelta

from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_cli.utils import time
from argus_plugins import argus_cli_module
from argus_plugins.cases import utils


GROUPINGS = {
    "year": lambda case: datetime.fromtimestamp(
        case["createdTimestamp"] / 1000
    ).strftime("%Y"),
    "month": lambda case: datetime.fromtimestamp(
        case["createdTimestamp"] / 1000
    ).strftime("%Y-%m"),
    "week": lambda case: datetime.fromtimestamp(
        case["createdTimestamp"] / 1000
    ).strftime("%Y-W%W"),
    "day": lambda case: datetime.fromtimestamp(
        case["createdTimestamp"] / 1000
    ).strftime("%Y-%m-%d"),
    "user": lambda case: case["createdByUser"]["userName"]
    if "createdByUser" in case
    else "N/A",
}

MISSING_GROUPER = {
    "year": lambda start, end: (
        start + relativedelta(years=i)
        for i in range(0, relativedelta(end, start).years)
    ),
    "month": lambda start, end: (
        start + relativedelta(months=i)
        for i in range(0, relativedelta(end, start).months)
    ),
    "week": lambda start, end: (
        start + timedelta(weeks=i) for i in range(0, (end - start).days // 7)
    ),
    "day": lambda start, end: (
        start + timedelta(days=1) for i in range(0, (end - start).days)
    ),
    # There is nothing to fill in for users, the users that are there are the
    # ones that posted. There are no "in between" users. An alternative could
    # be to add previously
    "user": lambda start, end: iter(()),
}


def add_missing_groups(
    existing_groups: dict, group_by: str, start_time: datetime, end_time: datetime
):
    """Adds a missing group to a set of grouped cases.

    I.e., given the following set of groups, with a start of 2019 and end of 2022:
        2019, 2020, 2022
    It will fill the missing groups:
        2019, 2020, 2021, 2022
    """
    periods = MISSING_GROUPER[group_by]
    group_getter = GROUPINGS[group_by]

    start_time = datetime.fromtimestamp(start_time / 1000)
    end_time = datetime.fromtimestamp(end_time / 1000)
    groups = existing_groups.copy()

    for period in periods(start_time, end_time):
        # GROUPINGS expects an argus style timestamp, and its easier to do it
        # like this than changing how GROUPINGS work.
        iso_time_string = group_getter({"createdTimestamp": period.timestamp() * 1000})
        if iso_time_string in groups:
            continue

        groups[iso_time_string] = []

    return groups


def group_cases(cases: list, group_by: str) -> dict:
    groups = {}
    grouping = GROUPINGS[group_by]

    for case in cases:
        group = grouping(case)
        if group not in groups:
            groups[group] = []
        groups[group].append(case)

    return groups


def sort_cases(cases: list) -> dict:
    """Sorts cases based on their priority"""
    sorted = {priority: [] for priority in utils.PRIORITIES}
    for case in cases:
        sorted[case["priority"]].append(case)

    return sorted


def print_statistics(cases: dict, grouping: str):
    """Prints the statistics as a CSV"""
    headers = [grouping]
    headers.extend(utils.PRIORITIES)
    print(",".join(headers))

    for group_id, group in sorted(cases.items()):
        stats = [str(len(group[priority])) for priority in utils.PRIORITIES]
        print("{group},{stats}".format(group=group_id, stats=",".join(stats)))


@register_command(extending="cases", module=argus_cli_module)
def statistics(
    start: time.date_or_relative,
    end: time.date_or_relative,
    customer: utils.get_customer_id = None,
    group_by: GROUPINGS = "week",
    case_type: utils.CASE_TYPES = None,
    category: str = None,
    service: str = None,
    status: utils.STATUSES = None,
    include_empty_groups: bool = False,
):
    """Shows the statistics about how many cases there are of each severity.

    :param start: The time of the first case
    :param end: The time of the last case
    :param list customer: The customer to search for
    :param group_by: How the info will be grouped
    :param list case_type: Type to search for
    :param list category: The category to search for
    :param list service: The service to search for
    :param list status: Status to search for
    :param include_empty_groups: When set, includes periods without any cases.
    :return:
    """
    log.info("Getting cases...")
    cases = advanced_case_search(
        limit=0,
        startTimestamp=start,
        endTimestamp=end,
        timeFieldStrategy=["createdTimestamp"],
        customerID=customer,
        type=case_type,
        category=category,
        service=service,
        status=status,
    )["data"]

    log.info("Grouping cases...")
    cases = group_cases(cases, group_by)

    if include_empty_groups:
        cases = add_missing_groups(cases, group_by, start, end)

    log.info("Sorting cases...")
    for group_id, group in cases.items():
        cases[group_id] = sort_cases(group)

    log.info("Printing statistics...")
    print_statistics(cases, group_by)



from copy import copy
from csv import DictReader, reader
from json import dumps
from pathlib import Path
from sys import exit
from typing import Callable, Dict, Optional
from urllib.parse import quote_plus

from argus_api.api.customers.v2.customer import get_customer
from argus_api.api.datastores.v1.store.store import (
    delete_data_store_entries,
    get_entries_from_store,
    put_data_store_entries,
)

from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_plugins import argus_cli_module


def get_buffered_data(func: Callable) -> Callable:
    """

    :param func: The getter function for the data to be buffered.
    :return: A generator with the data, where the data is the ["data"] parameter of the response.
    """

    def ann(*args, **kwargs):
        buffer_size = 200

        offset = 0
        response = func(*args, limit=buffer_size, offset=offset, **kwargs)
        while offset < response["count"]:
            yield from response["data"]
            offset += buffer_size

            response = func(*args, limit=buffer_size, offset=offset, **kwargs)

    return ann


def set_data_with_buffer(func: Callable, buffered_argument: str) -> Callable:
    """
    Decorator to send data in multiple batches.

    Large requests might be eaten by a proxy or the service itself.
    Because of this, we're setting the batch size to 4k. This is half of the
    max size of a URL according to the HTTP RFC, and should be well within a
    safe limit to assure that our request is delivered.

    :param func: The function to buffer.
    :param buffered_argument: The kwarg that is going to be split into
                              multiple requests. The key that this is
                              referencing has to be an iterable.
    :return: The decorated function.
    """

    def buffered_arguments(args, kwargs):
        buffer = kwargs.pop(buffered_argument)
        buffer_size = len(buffer)

        if any(len(quote_plus(str(item))) > 4000 for item in buffer):
            raise ValueError(
                "One of the arguments to {} is over the 4K "
                "characters. This exceeds the maximum limit and "
                "is not allowed."
            )

        offset = 0
        while offset < buffer_size:
            current_end = offset + 1
            while (
                sum(
                    len(str(s).encode("unicode_escape"))
                    for s in buffer[offset : current_end + 1]
                )
                < 4000
                and buffer_size >= current_end + 1 > offset
            ):
                current_end += 1

            kwargs[buffered_argument] = buffer[offset:current_end]
            offset = current_end

            yield args, kwargs

    def buffered_function(*args, **kwargs):
        response = None
        for args, kwargs in buffered_arguments(args, kwargs):
            response = func(*args, **kwargs)

        return response

    return buffered_function


def datastore_data(
    data: str,
    csv_key: str = "key",
    lowercase_key: bool = False,
    lowercase_value: bool = False,
    default_value: str = "N/A",
) -> Dict[str, str]:
    """Turns a CSV string into key:value pairs."""

    try:
        with open(data, "r") as fp:
            data = fp.readlines()
    except IOError:
        data = data.split("\n")

    # Remove any comments
    data = filter(
        lambda line: not line.startswith("#") and len(line) and not line.isspace(), data
    )
    csv_reader = reader(data)

    entries = {}
    for row in csv_reader:
        key = row[0].strip()

        if lowercase_key:
            key = key.lower()

        value = (
            format_value(row[1], lowercase_value, default_value)
            if len(row) > 1
            else format_value(default_value, lowercase_value, default_value)
        )

        if key in entries:
            log.warning(
                f'The key "{key}" exists multiple times in the data. Overriding with new value'
            )
        if key == csv_key:
            log.warning(
                f"You appear to be using an unformatted CSV file without using the --csv-json argument "
                f"""({'{'}"{key}": "{value}"{'}'} will be added to the datastore)"""
            )

        entries[key] = value

    return entries


def csv_to_json(
    csv_file_path: Path,
    csv_key: str = "key",
    lowercase_key: bool = False,
    lowercase_value: bool = False,
) -> dict:
    """Convert CSV file WITH headers to dict where the CSV column
    specified in the `csv_key` parameter is left intact, and the rest
    of the columns are converted to a stringified dict with the column name as key.

    :param csv_file_path: Filepath of the CSV file.
    :param csv_key: Column to use as keys in the dict.
    """

    csv_data = {}

    with open(csv_file_path, "r") as csv_in:
        # DictReader formats each row from the CSV file to {csv_key: row_key, header1: row_val1, header2: row_val2, ...}
        for row in DictReader(csv_in):
            # If `csv_key` does not exist in the CSV headers, raise ValueError
            if csv_key not in row.keys():
                raise ValueError(f"No column `{csv_key}` in CSV file")

            # Check if the row is commented out
            if list(row.values())[0].startswith("#"):
                log.info(f"Skipping key starting with #: {row}")
                continue

            # Convert row to format:  row_key: "{'header1': 'row_val1', 'header2': 'row_val2', ...}"
            value = copy(row)
            del value[csv_key]

            if lowercase_value:
                value = {header: v.lower() for header, v in value.items()}

            key = row[csv_key]

            if lowercase_key:
                key = key.lower()

            # Serialize the value to a JSON formatted string
            csv_data[key] = dumps(value)

    return csv_data


def validate_csv_options(csv_json: bool, csv_key: str) -> None:
    """Verify that the csv_json option is True if the csv_key option is provided.

    :param csv_json: Flag to enable formatting of the input CSV file.
    :param csv_key: String corresponding to a column in the CSV file to use as keys in the Argus datastore.
    :raises AttributeError: When csv_json is False and the csv_key option is provided with a different value than `key`.
    """

    if csv_key != "key" and not csv_json:
        raise AttributeError(
            "--csv-key is only valid when the --csv-json option is also provided."
        )


def validate_data_format(data: str) -> None:
    """Verify that the data argument is a valid CSV file or a dict.

    :param data: String corresponding to a CSV file, or a python dict.
    :raises FileNotFoundError: If data is not a dict and does not point to a file.
    """

    if isinstance(data, dict):
        return

    if not Path(data).is_file():
        raise FileNotFoundError(f"File not found: '{data}'")


@register_command(extending="datastores", module=argus_cli_module)
def delete(datastore: str, keys: list, customer: str = None) -> None:
    """Deletes given entries from the datastore.

    :param datastore: The datastore to modify.
    :param customer: The customer to affect.
    :param keys: Keys to delete. A file can be provided with the @-prefix
                 (eg. @/tmp/datastore_delete.txt).
    """

    if customer:
        customer_id = get_customer(idOrShortName=customer.lower())["data"]["id"]
    else:
        customer_id = None

    buffered_delete = set_data_with_buffer(
        delete_data_store_entries, buffered_argument="key"
    )
    buffered_delete(dataStore=datastore, customerID=customer_id, key=keys)

    print(f"Successfully deleted {len(keys)} entries")


def format_value(
    value: Optional[str], lowercase_value: bool, default_value: str
) -> str:
    """Handle default values, so the value is always stripped and lowecased (if flag is set)

    :param value: Value
    :param lowercase_value: Return lowercase of value
    :param default_value: Return this value if value is not truthy
    :return: formatted value
    """
    if not value:
        value = default_value

    if lowercase_value:
        return value.lower().strip()

    return value.strip()


@register_command(extending="datastores", module=argus_cli_module)
def update(
    datastore: str,
    data: str,
    customer: str = None,
    default_value: str = "N/A",
    csv_json: bool = False,
    csv_key: str = "key",
    lowercase_key: bool = False,
    lowercase_value: bool = False,
):
    """Adds or updates entries from the data.

    :param datastore: The Argus datastore.
    :param data: CSV file/python dict. The Argus datastore will be updated to match the data in this file/dict.
    :param customer: Specifies which customer to update entries for in the datastore.
    :param default_value: Value inserted for missing values in the Argus datastore.
    :param csv_json: Flag which when set makes the script convert input CSV file to a dict using column specified in
    `csv_key` as keys and residual fields in each row as stringified values.
    :param csv_key: String corresponding to the column in the CSV file to use as keys in the Argus datastore.
    :param lowercase_key: Flag which when set will lowercase the key when reading from CSV
    :param lowercase_value: Flag which when set will lowercase the values when reading from CSV
    """

    validate_csv_options(csv_json=csv_json, csv_key=csv_key)
    validate_data_format(data=data)

    if csv_json:
        try:
            data = csv_to_json(
                csv_file_path=Path(data),
                csv_key=csv_key,
                lowercase_key=lowercase_key,
                lowercase_value=lowercase_value,
            )
        except (ValueError, FileNotFoundError) as e:
            log.exception(e)
            exit(1)
    elif isinstance(data, str):
        # Only call datastore_data if update was invoked from the cli (not from the sync method)
        data = datastore_data(
            data,
            csv_key,
            lowercase_key=lowercase_key,
            lowercase_value=lowercase_value,
            default_value=default_value,
        )

    if customer:
        customer_id = get_customer(idOrShortName=customer.lower())["data"]["id"]
    else:
        customer_id = None

    entries = [
        {
            csv_key: key.lower() if lowercase_key else key,
            "value": format_value(value, lowercase_value, default_value),
        }
        for key, value in data.items()
    ]
    buffered_put = set_data_with_buffer(
        put_data_store_entries, buffered_argument="entries"
    )
    response = buffered_put(
        dataStore=datastore, entries=entries, customerID=customer_id
    )

    print(f"Successfully updated {len(entries)} entries")


@register_command(extending="datastores", module=argus_cli_module)
def sync(
    datastore: str,
    data: str,
    customer: str = None,
    default_value: str = "N/A",
    csv_json: bool = False,
    csv_key: str = "key",
    lowercase_key: bool = False,
    lowercase_value: bool = False,
):
    """Makes sure the datastore is a 1:1 match with the given data (for a given customer, if any).

    :param datastore: The Argus datastore.
    :param data: CSV file. The Argus datastore will be updated to match the data in this file.
    :param customer: Specifies which customer to update entries for in the datastore.
    :param default_value: Value inserted for missing values in the Argus datastore.
    :param csv_json: Flag which when set makes the script convert input CSV file to a dict using column specified in
    `csv_key` as keys and residual fields in each row as stringified values.
    :param csv_key: String corresponding to the column in the CSV file to use as keys in the Argus datastore.
    :param lowercase_key: Flag which when set will lowercase the key when reading from CSV
    :param lowercase_value: Flag which when set will lowercase the values when reading from CSV
    """

    validate_csv_options(csv_json=csv_json, csv_key=csv_key)
    validate_data_format(data=data)

    if csv_json:
        try:
            data = csv_to_json(
                csv_file_path=Path(data),
                csv_key=csv_key,
                lowercase_key=lowercase_key,
                lowercase_value=lowercase_value,
            )
        except (ValueError, FileNotFoundError) as e:
            log.exception(e)
            exit(1)
    else:
        data = datastore_data(
            data,
            csv_key,
            lowercase_key=lowercase_key,
            lowercase_value=lowercase_value,
            default_value=default_value,
        )

    if customer:
        customer_id = get_customer(idOrShortName=customer.lower())["data"]["id"]
    else:
        customer_id = None

    delete_entries = []  # Items to delete
    update_entries = {}  # Items to update/add
    existing_entries = {}  # Entries that already exist

    fetched_entries = get_buffered_data(get_entries_from_store)(datastore)

    for entry in fetched_entries:
        if customer_id and entry["customer"]["id"] != customer_id:
            continue

        key = entry[csv_key]
        value = entry["value"] or None

        existing_entries[key] = value
        if key not in data:
            delete_entries.append(key)

    for key, value in data.items():
        if value != existing_entries.get(key):
            update_entries[key] = value

    # Delete entries before we update, since it looks like deletes are not
    # case sensitive, so we will drop entries if we are changing case on an entry
    if delete_entries:
        delete(datastore, delete_entries, customer)

    if update_entries:
        update(datastore, update_entries, customer)



from .remind import remind
from .statistics import statistics
from .basic import change_status
from .autocreate_case import autocreate
from .autocreate_yaml import autocreate_yaml
from .notification_check import notification_check



import json
from datetime import datetime

from argus_api.api.customernetworks.v1.network import update_customer_network
from argus_api.exceptions.http import ArgusException

from argus_cli.helpers import formatting
from argus_cli.helpers.formatting import success, failure

from argus_cli.helpers.log import log
from argus_plugins.customer_networks.helpers.diff import diff_table


def update(
    networks: list,
    customer: dict,
    as_json: bool = False,
    dry: bool = False,
    skip_confirmation: bool = False,
) -> list:
    """Updates existing networks

    :param list networks: List of CustomerNetworks
    :param bool as_json: Messages are returned as JSON
    :param bool dry: Whether any actual changes should be performed or not
    :param bool interactive: Ask the user to continue on each network?
    :return list[str]: Summary of messages produced
    """

    # Find all networks that have changed in the given list:
    dirty_networks = [n for n in networks if n.is_dirty()]

    # This will hold our messages (failure, success, etc), which we display at the end
    summary = []

    # If we're running in interactive mode, then print a difference
    # table for each network that will be changed, and ask the user
    # to confirm each individual change instead
    for index, network in enumerate(dirty_networks):
        # Clear terminal when interactive:
        formatting.clear()
        print(
            formatting.table(
                data=diff_table([network]),
                keys=("IP", "Description", "Zone", "Location", "Flags"),
                title="Changes to %s (%d of %d)"
                % (network.to_json()["networkAddress"], index + 1, len(networks)),
            )
        )

        # Ask the user to confirm the change unless this has
        # been overridden by skip_confirmation
        if skip_confirmation or formatting.ask("Accept changes?"):
            log.info("Updating networks for %s" % customer["name"])

            try:
                # Only perform API actions when not in dry mode
                if not dry:
                    # Update the network with its dirty
                    # (modified) fields before creating JSON
                    # from it
                    network.update(network._dirty)
                    update_data = {
                        key: value
                        for key, value in network.to_json().items()
                        # Allowed fields for update:
                        if key
                        in (
                            "networkID",
                            "location",
                            "description",
                            "flagsToEnable",
                            "flagsToDisable",
                            "zone",
                        )
                    }
                    update_customer_network(**update_data)
                # Add success message to summary report
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Network %s updated on Argus"
                            % (network.to_json()["networkAddress"]),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": True,
                            "created": False,
                            "failure": False,
                            "deleted": False,
                        }
                    )
                else:
                    summary.append(
                        success(
                            "Network %s updated on Argus"
                            % (network.to_json()["networkAddress"])
                        )
                    )
                    log.info(summary[-1])
            except ArgusException as error:
                if as_json:
                    summary.append(
                        {
                            "timestamp": datetime.now().isoformat(),
                            "message": "Failed to update network %s on Argus: %s"
                            % (network.to_json()["networkAddress"], error),
                            "name": network["description"],
                            "ip": network.to_json()["networkAddress"],
                            "body": network.to_json(),
                            "updated": True,
                            "created": False,
                            "failure": False,
                            "deleted": False,
                        }
                    )
                else:
                    summary.append(
                        failure(
                            "Failed to update network %s on Argus: %s"
                            % (network.to_json()["networkAddress"], error)
                        )
                    )
                    log.error(summary[-1])
    return summary



import json
from functools import partial
from argus_cli.plugin import register_command
from argus_cli.helpers import formatting

from argus_api.api.customers.v2.customer import get_customer
from argus_api.api.customernetworks.v1.network import search_customer_network
from argus_plugins import argus_cli_module

from .customer_network import CustomerNetwork
from .helpers import interactive as interactive_action, batch as batch, diff


@register_command(extending="customer-networks", module=argus_cli_module)
def update(
    customer: str,
    file: str,
    format: str = "address,description",
    first_line_defines_format: bool = False,
    always_yes: bool = False,
    replace: bool = False,
    interactive: bool = False,
    delete_missing: bool = False,
    splunk: bool = False,
    output: str = "",
    delimiter: str = ",",
    dry: bool = False,
) -> None:
    """
    Update customer networks from CSV file.

    NOTICE: Any customer fields in the CSV file itself will be ignored. The customer specified in the CLI will be used
    regardless.

    :param customer: Customer shortname or ID
    :param file: Path to CSV file
    :param format: CSV format (available fields: address (alias Network), description, subnet_mask (optional, can be part of the address), zone, location (alias LocationShortName), Flags)
    :param first_line_defines_format: If set, --format will be ignored and the CSV file is assumed to define the fields on the first line. Must include the fields mentioned in --format.
    :param always_yes: Skip confirmation on updates (WARNING: If run with Replace, this will always replace existing networks, if run with Delete, networks not found in CSV will always be deleted from Argus! Use with caution.)
    :param replace: Update networks that already exist with new information
    :param delete_missing: Remove networks on Argus if they dont exist in the file (update, create from file and remove from argus)
    :param delimiter: CSV delimiter character
    :param interactive: Performs updates interactively and requires user to confirm each change
    :param dry: Performs a dry run - no updates will be sent to Argus
    :param splunk: Suppresses log and writes JSON output instead
    :param output: Used in conjunction with the splunk flag to write JSON to file
    :alias format: F
    :alias first_line_defines_format: L
    :alias always_yes: y
    :alias replace: R
    :alias delete_missing: X
    :alias delimiter: t
    :alias interactive: i
    :alias dry: D

    """
    # Find customer by name
    customer = get_customer(idOrShortName=customer)

    if not customer:
        raise LookupError(
            "No customer has been selected. Customer is required to edit customer networks."
        )

    customer = customer["data"]

    # Otherwise select the only customer
    # Fetch all networks for customer
    existing_networks = search_customer_network(
        customerID=[customer["id"]],
        # Setting limit to -1 does nothing, and 0 sets default limit to 25
        limit=0,
    )

    _from_json = partial(CustomerNetwork.from_json, customer=customer)
    # Convert them to CustomerNetwork objects
    existing_networks = list(map(_from_json, existing_networks["data"]))

    changes = diff.diff_file_vs_argus(
        network_list=CustomerNetwork.from_csv(
            file,
            header_format=None if first_line_defines_format else format.split(","),
            headers_on_first_line=first_line_defines_format,
            delimiter=delimiter,
            customer=customer,
        ),
        existing_networks=existing_networks,
    )

    messages = []

    if interactive:
        create = interactive_action.create
        update = interactive_action.update
        destroy = interactive_action.destroy
    else:
        create = batch.create
        update = batch.update
        destroy = batch.destroy

    if replace and len(changes["CHANGED_IN_FILE"]):
        messages += update(
            changes["CHANGED_IN_FILE"],
            customer,
            dry=dry,
            as_json=splunk,
            skip_confirmation=always_yes,
        )

    if len(changes["NOT_ON_SERVER"]):
        messages += create(
            changes["NOT_ON_SERVER"],
            customer,
            dry=dry,
            as_json=splunk,
            skip_confirmation=always_yes,
        )

    if delete_missing and len(changes["NOT_IN_FILE"]):
        messages += destroy(
            changes["NOT_IN_FILE"],
            customer,
            dry=dry,
            as_json=splunk,
            skip_confirmation=always_yes,
        )

    # Print summary
    if len(messages):
        if splunk:
            if output:
                with open(output, "w") as json_file:
                    json_file.write(json.dumps(messages))
            else:
                print(json.dumps(messages, indent=4))
        else:
            formatting.clear()
            print(
                formatting.table(
                    data=[{"message": msg} for msg in messages],
                    title="Summary",
                    keys=["message"],
                )
            )
    else:
        print("No action required.")



from argus_api.api.customers.v2.customer import get_customer

# These are copy pasted from the argus documentation.
# It's recommended to view that if you want to get an overview of flags, etc.
GROUPINGS = [
    "ATTACK_IDENTIFIER",
    "ATTACK_CATEGORY",
    "ALARM",
    "SENSOR",
    "LOCATION",
    "EVENT_SEVERITY",
    "SOURCEIP",
    "DESTINATIONIP",
    "PROTOCOL",
    "CUSTOMER",
    "SOURCE_GEOIP_LOCATION",
    "SOURCE_GEOIP_COUNTRY",
    "DESTINATION_GEOIP_LOCATION",
    "DESTINATION_GEOIP_COUNTRY",
]

#: Valid event flags
FLAGS = [
    "ESTABLISHED",
    "BLOCKED",
    "SNAPSHOT",
    "FINALIZED",
    "SOURCE_IS_CUSTOMERNET",
    "DESTINATION_IS_CUSTOMERNET",
    "SOURCE_IS_PARTIAL_CUSTOMERNET",
    "DESTINATION_IS_PARTIAL_CUSTOMERNET",
    "INTRUDER_IS_CUSTOMERNET",
    "VICTIM_IS_CUSTOMERNET",
    "INTRUDER_IS_PARTIAL_CUSTOMERNET",
    "VICTIM_IS_PARTIAL_CUSTOMERNET",
    "PARTIALLY_BLOCKED",
    "FALSE_POSITIVE",
    "NOT_A_THREAT",
    "TUNING_CANDIDATE",
    "NOTIFIED",
    "PARTIALLY_NOTIFIED",
    "NOTIFIED_UNPUBLISHED",
    "NOTIFIED_DELETED",
    "FOLLOWUP",
    "IDENTIFIED_THREAT",
    "THREAT_CANDIDATE",
    "ACKNOWLEDGED",
    "PARTIALLY_ACKNOWLEDGED",
    "SEVERITY_ADJUSTED",
    "COMMENTED",
    "FILTERED",
    "CHECKED",
    "INCOMPLETE_DETAILS",
    "AGGREGATED_BASE_EVENT",
    "REMOTE_STORAGE",
    "CUSTOM_SOURCE_AGGREGATION",
    "CUSTOM_DESTINATION_AGGREGATION",
    "CUSTOM_INTRUDER_AGGREGATION",
    "CUSTOM_VICTIM_AGGREGATION",
    "HAS_DETAILS",
    "HAS_PAYLOAD",
    "HAS_PCAP",
    "ASSOCIATED_TO_CASE_BY_FILTER",
    "SEVERITY_INCREASED_BY_FILTER",
    "SEVERITY_REDUCED_BY_FILTER",
    "CREATED_BY_ANALYSIS_FILTER",
    "EXTEND_EVENT_TTL",
    "INITIAL_TUNING",
    "POST_ANALYSIS",
    "PARTIAL_SSL_TERMINATED",
    "SSL_TERMINATED",
    "AUTO_REPORT",
    "MISSING_TIMESTAMP",
    "CLOCK_OUT_OF_SYNC",
    "DROP_ANALYSIS",
    "ESCALATED_BY_REPUTATION",
    "HAS_SAMPLE",
    "STORE_EVENT",
    "STORE_AGGREGATED",
    "SOURCE_IS_MANAGED_BY_SOC",
    "DESTINATION_IS_MANAGED_BY_SOC",
    "HANDLED_BY_ANALYST",
    "SLA_VIOLATION",
    "PAYLOAD_TRUNCATED",
    "HAS_STRING_PAYLOAD",
    "REASSESSED",
]

#: Valid event severity values
SEVERITIES = ["low", "medium", "high", "critical"]


def get_customer_id(name: str) -> int:
    """Gets a customer's ID from their name

    :param name: The name of the customer
    """
    customers = get_customer(idOrShortName=name.lower())["data"]
    customer_id = customers[
        "id"
    ]  # This might get the wrong customer if there are more with the same name?
    return customer_id



from argus_cli.plugin import register_command
from argus_cli.helpers import formatting

from argus_api.api.customers.v2.customer import get_customer
from argus_api.api.customernetworks.v1.network import get_customer_networks
from argus_api.api.sensors.v1.location import list_locations as _list_locations
from argus_plugins import argus_cli_module


@register_command(extending="customer-networks", module=argus_cli_module)
def list(customer: str) -> None:
    """
    This function handles the subcommand "show", and only fetches and displays a list of existing networks

    :param customer: Customer shortname or ID
    """

    # Find customer by name
    customer = get_customer(idOrShortName=customer)

    if not customer:
        raise LookupError(
            "No customer has been selected. Customer is required to edit customer networks."
        )

    # Otherwise select the only customer
    networks = get_customer_networks(customerID=[customer["data"]["id"]], limit=0)

    def format_column(value: dict, key: str):
        if key == "flags" and value:
            return ",".join(value)
        elif key == "networkAddress":
            return "%s/%s" % (value["address"], value["maskBits"])
        elif key == "location":
            return value["name"]
        else:
            return value or "-"

    print(
        formatting.table(
            data=networks["data"],
            keys=["networkAddress", "description", "zone", "location", "flags"],
            format=format_column,
            title="Networks for %s" % customer["data"]["name"],
        )
    )


@register_command(extending="customer-networks", module=argus_cli_module)
def list_locations(customer: str) -> None:
    """
    This function handles the subcommand "show", and only fetches and displays a list of existing networks

    :param customer: Customer whose locations to search (shortname or ID).
    """

    # Find customer by name
    customer = get_customer(idOrShortName=customer)["data"]

    locations = _list_locations(customerID=customer["id"], limit=0)["data"]

    def format_column(value: dict, key: str):
        if key == "flags" and value:
            return ",".join(value)
        elif key == "networkAddress":
            return "%s/%s" % (value["address"], value["maskBits"])
        elif key == "location":
            return value["name"]
        else:
            return value or "-"

    print(
        formatting.table(
            data=locations,
            keys=["id", "shortName", "zone", "name", "flags"],
            format=format_column,
            title="Locations",
        )
    )



from argus_cli.plugin import create_module, import_external_plugins
from argus_plugins.api_provider import argus_api
from argus_cli import __version__, argus_api_version, internal_plugins_version

cli_versions = f"argus-toolbelt: {__version__}, argus-api: {argus_api_version}, argus-cli-internal-plugins: {internal_plugins_version}"

__all__ = [
    "assets",
    "cases",
    "customer_networks",
    "datastore",
    "events",
    "reports",
]

# Initialize the toolbelt with the framework.
argus_cli_module = create_module(providers=[argus_api])
argus_cli_module.argument_parser.add_argument(
    "--version", action="version", version=cli_versions
)



try:
    from importlib.metadata import version, PackageNotFoundError
except ImportError:
    from importlib_metadata import version, PackageNotFoundError

__version__ = version("argus-toolbelt")

argus_api_version = version("argus-api")
try:
    internal_plugins_version = version("argus-cli-internal-plugins")
except PackageNotFoundError:
    internal_plugins_version = "not installed"

from .plugin import register_command, register_provider, run



from pathlib import Path

import yaml
from tqdm import tqdm

from argus_api.api.events.v1.aggregated import find_aggregated_events
from argus_cli.plugin import register_command
from argus_cli.helpers.log import log
from argus_cli.helpers.pagination import offset_paginated
from argus_cli.utils.arguments import organize_parameters
from argus_cli.utils.formatting import get_data_formatter
from argus_cli.utils.time import timestamp_to_date, date_or_relative
from argus_plugins import argus_cli_module

from argus_plugins.cases.utils import get_customer_id
from argus_plugins.events import utils
from argus_plugins.events.search_model import SearchOptions
from argus_plugins.enrichments import EVENTS_ENRICHMENTS, enrich_data

#: hard limit on maximum results the backend can return
SEARCH_RESULTS_MAX = 100000

argument_api_name_map = {
    "customer": "customerID",
    "flag": "includeFlags",
    "alarm": "alarmID",
    "ip": "ip",
    "source_ip": "sourceIP",
    "destination_ip": "destinationIP",
    "signature": "signature",
    "properties": "properties",
    "attack_category_id": "attackCategoryID",
}


def rename_variables(parameters: dict) -> tuple:
    """Renames input parameters to a more programmable format

    :param parameters: Parameters to convert
    :return: Include and exclude parameters in dicts
    """
    include = {
        argument_api_name_map[key.replace("include_", "")]: value
        for key, value in parameters.items()
        if key.startswith("include_") and value
    }
    exclude = {
        argument_api_name_map[key.replace("exclude_", "")]: value
        for key, value in parameters.items()
        if key.startswith("exclude_") and value
    }

    if "includeFlags" in exclude.keys():
        # Hack because include/exclude flag differs from all other parameters
        include["excludeFlags"] = exclude.pop("includeFlags")

    return include, exclude


def format_event(event: dict) -> dict:
    """Replaces some keys from the event to human readable format

    The following keys are modified:
        r"*timestamp*": Timestamp to ISO8601 date-time format.

    :param event: Event to modify
    :returns: Modified event
    """
    for time_key in filter(lambda key: "timestamp" in key.lower(), event.keys()):
        event[time_key] = timestamp_to_date(event[time_key] / 1000)

    return event


@register_command(extending="events", module=argus_cli_module)
def search(
    start: date_or_relative,
    end: date_or_relative,
    format: get_data_formatter = "csv",
    include_customer: get_customer_id = None,
    exclude_customer: get_customer_id = None,
    include_flag: utils.FLAGS = None,
    exclude_flag: utils.FLAGS = None,
    include_alarm: int = None,
    exclude_alarm: int = None,
    include_ip: str = None,
    exclude_ip: str = None,
    include_source_ip: str = None,
    exclude_source_ip: str = None,
    include_destination_ip: str = None,
    exclude_destination_ip: str = None,
    include_signature: str = None,
    exclude_signature: str = None,
    include_properties: dict = None,
    exclude_properties: dict = None,
    include_attack_category_id: int = None,
    exclude_attack_category_id: int = None,
    exact_match_properties: bool = True,
    min_severity: utils.SEVERITIES = "high",
    min_count: int = None,
    source_ip_min_bits: int = None,
    destination_ip_min_bits: int = None,
    enrich: EVENTS_ENRICHMENTS = None,
    limit: int = 500,
    paginate: bool = True,
    progress: bool = False,
    print_events: bool = True,
):
    """Searches for events in argus.

    Include and exclude parameters are mimics the aggregated event search endpoint,
    and can be found in the api documentation.

    Example use:
        argus-cli events search "2 days ago" "now"
            --format json
            --include-properties '{"argus.timestamp": 9001, "uhost": "host.example.com"}'

    :param start: Start time of the search in ISO8016 format (Example: 2018-01-01)
    :param end: End time of the search in ISO8016 format (Example: 2018-01-01)
    :param format: Follows the normal str.format() syntax.
    :param list include_customer: Customer(s) to include
    :param list exclude_customer: Customer(s) to exclude
    :param list include_flag: Flag(s) to include
    :param list exclude_flag: Flag(s) to exclude
    :param list include_alarm: Alarm ID(s) to include
    :param list exclude_alarm: Alarm ID(s) to exclude
    :param list include_ip: IP(s) to include
    :param list exclude_ip: IP(s) to exclude
    :param list include_source_ip: Source IP(s) to include
    :param list exclude_source_ip: Source IP(s) to exclude
    :param list include_destination_ip: Destination IP(s) to include
    :param list exclude_destination_ip: Destination IP(s) to exclude
    :param list include_signature: Signature(s) to include
    :param list exclude_signature: Signature(s) to exclude
    :param include_properties: Properties to include
    :param exclude_properties: Properties to exclude
    :param list include_attack_category_id: attack category ID(s) to include
    :param list exclude_attack_category_id: attack category ID(s) to exclude
    :param exact_match_properties: Exact match of properties specified at the top level
    :param min_severity: Minimum severity of events
    :param min_count: Minimum aggregated count of events
    :param source_ip_min_bits: Do not include source CIDR\-networks with wider mask than this
    :param destination_ip_min_bits: Do not include destination CIDR\-networks with wider mask than this
    :param list enrich: optional enrichments to perform on the events
    :param limit: number of items to fetch per request ("page size") - default 500
    :param paginate: if set, disable pagination (only the number of results set by --limit will be fetched)
    :param progress: if set, display a progress bar
    :param print_events: set by default. if unset, events will not be printed out.

    :alias format: format-string
    """
    log.info("Organizing parameters...")
    include, exclude = rename_variables(locals())
    parameters = organize_parameters(include, exclude)
    if limit > 1000:
        log.warning(
            "limit is higher that 1000, which can negatively impact the service. Consider setting a lower limit."
        )

    log.info("Getting events from Argus...")
    events = []
    pages = offset_paginated(find_aggregated_events)(
        limit=limit,
        startTimestamp=start,
        endTimestamp=end,
        minSeverity=min_severity,
        minCount=min_count,
        sourceIPMinBits=source_ip_min_bits,
        destinationIPMinBits=destination_ip_min_bits,
        exactMatchProperties=exact_match_properties,
        **parameters,
    )
    if not paginate:
        pages = [pages._first_page]
    progress_bar = None
    for page in pages:
        if page["count"] > SEARCH_RESULTS_MAX:
            log.error(
                f"this search would return more than {SEARCH_RESULTS_MAX} items "
                f"({page['count']}), which is not supported. "
                f"Please narrow your search."
            )
            exit(1)
        if not progress_bar:
            # initialize the progress bar only once we have a count to avoid displaying
            # an empty bar while waiting for the first batch
            progress_bar = tqdm(
                disable=not progress,
                dynamic_ncols=True,
                total=page["count"],
                initial=page["size"],
            )
        else:
            # refresh the total, since data is dynamic
            progress_bar.total = page["count"]
            progress_bar.refresh()
            # update progress manually
            progress_bar.update(page["size"])
        events.extend(page["data"])
    if progress_bar:
        # can be uninitialized if the result set is empty
        progress_bar.close()

    if not events:
        log.info("There are no events in the search.")
        if not print_events:
            return format([])

    if enrich is not None:
        log.debug(f"Enriching events with {' '.join(enrich)}")
        enrich_data(events, enrich)

    log.info("Formatting output to human-readable format...")
    for event in events:
        event = format_event(event)

    formatted = format(events)

    if print_events:
        log.info("Printing output...")
        print(formatted)
    else:
        log.info(f"found {len(events)} events")
        return formatted


@register_command(extending="events", module=argus_cli_module)
def search_yaml(
    config: Path,
    start: date_or_relative,
    end: date_or_relative,
    format: get_data_formatter = "csv",
    limit: int = 500,
    paginate: bool = True,
    progress: bool = False,
):
    """Run a search described in configuration file.

    Example configuration:

      customer:
        include:
          - "mnemonic"
      ip:
        exclude:
          - "127.0.0.1"
        min_severity: "high"
        properties:
          include:
            "example.property": "value"


    :param config: path to the search configuration file.
    :param start: start time of the search in ISO8016 format (Example: 2018-01-01)
    :param end: end time of the search in ISO8016 format (Example: 2018-01-01)
    :param format: follows the normal str.format() syntax.
    :param limit: number of items to fetch per request ("page size") - default 500
    :param paginate: if set, disable pagination (only the number of results set by --limit will be fetched)
    :param progress: if set, display a progress bar
    """
    settings_raw = yaml.safe_load(config.expanduser().read_text())
    settings = SearchOptions(**settings_raw)
    args = settings.to_search_args()
    search(
        start=start,
        end=end,
        format=format,
        limit=limit,
        paginate=paginate,
        progress=progress,
        **args,
    )



"""Constants and common functions related to reports"""

PERIOD = (
    "day",
    "week",
    "custom",
    "month",
    "quarter",
    "year",
    "last_hour",
    "last_24hours",
    "last_7days",
)
STATUS = ("new", "assigned", "in_progress", "completed")



from argus_cli.plugin import register_command
from argus_cli.utils.time import date_time_to_timestamp
from argus_cli.utils.formatting import formatted_string
from argus_api.api.reports.v1.search import simplified_search_reports
from argus_plugins import argus_cli_module

from argus_plugins.cases.utils import get_customer_id
from argus_plugins.reports import PERIOD


def _sort_reports(reports: list) -> tuple:
    """Sorts the reports into published and not published"""
    published = []
    not_published = []

    for report in reports:
        if report["status"] == "COMPLETED":
            published.append(report)
        else:
            not_published.append(report)

    return published, not_published


def _print_results(published: list, not_published: list, format_string: str) -> None:
    """Prints the results in two lists"""
    output = (
        "Not Published\n-------------\n"
        + "{not_published}\n"
        + "\n"
        + "Published\n-------------\n"
        + "{published}"
    )

    output = output.format(
        not_published=formatted_string(not_published, format_string),
        published=formatted_string(published, format_string),
    )

    print(output)


@register_command("status", "reports", module=argus_cli_module)
def report_status(
    period: PERIOD,
    customer: get_customer_id = None,
    start: date_time_to_timestamp = None,
    end: date_time_to_timestamp = None,
    format_string: str = "{customer[name]} ({customer[shortName]}) - {description}",
):
    """Show what reports have been published and not published

    :param period: The period to view
    :param list customer: A set of customers to include in the search
    :param start: The start of the period (if in custom mode)
    :param end: The end of the period (if in custom mode)
    :param format_string: How each customer will be displayed (Used like str.format)
    """
    if period == "custom" and (not start or not end):
        print("A period has to be defined when having a custom period")
        exit(1)

    reports = simplified_search_reports(
        limit=0,
        startTime=start,
        endTime=end,
        period=period.upper(),
        customerID=customer,
    )["data"]

    published, not_published = _sort_reports(reports)
    _print_results(published, not_published, format_string)



from argus_cli.helpers import formatting


def diff_table(network_list: list, show_only_changes: bool = False) -> list:
    """Creates a dict with each network appearing twice, first
    with the existing data, in red, then with the changed data,
    in green, to show the diff before creating / updating / destroying

    :param bool show_only_changes: Only show the green row with changes (used for create action)
    :param list network_list: List of CustomerNetworks
    :returns: List with two rows for each network, one in green and one in red
    :returns: List[Dict[str, str]]
    """

    data = []
    for network in network_list:
        json = network.to_json()
        # Clean (before change) fields:
        if not show_only_changes:
            data.append(
                {
                    "IP": formatting.red("- " + json["networkAddress"]),
                    "Description": formatting.red(network["description"]),
                    "Zone": formatting.red(
                        network["zone"] if "zone" in network else ""
                    ),
                    "Location": formatting.red(
                        network["location"] if "location" in network else ""
                    ),
                    "Flags": formatting.red(", ".join(network["flags"])),
                }
            )

        json.update(network._dirty)
        # Changed fields:
        data.append(
            {
                "IP": formatting.green("+ " + json["networkAddress"]),
                "Description": formatting.green(json["description"]),
                "Zone": formatting.green(network["zone"] if "zone" in network else ""),
                "Location": formatting.green(
                    network["location"] if "location" in network else ""
                ),
                "Flags": formatting.green(", ".join(network.expected_flags())),
            }
        )
    return data


def diff_file_vs_argus(
    network_list: list,
    existing_networks: list,
) -> dict:
    """Compares a list of networks from a file to the networks on Argus for the customer,
    and returns a dict with keys: NOT_IN_FILE, CHANGED_IN_FILE, NOT_ON_SERVER, UNMODIFIED

    :param list network_list: Networks in file
    :param list existing_networks: Networks on Argus
    :param str api_key: API key to perform requests with
    :returns: Dict[str, list]
    """
    diff = {
        "NOT_IN_FILE": [],
        "CHANGED_IN_FILE": [],
        "NOT_ON_SERVER": [],
        "UNMODIFIED": [],
    }
    for network in existing_networks:
        # Check all existing networks
        if all((network not in network_on_file for network_on_file in network_list)):
            diff["NOT_IN_FILE"].append(network)
    for network in network_list:
        if any(
            (
                network_on_server.merge(network)
                for network_on_server in existing_networks
                if network in network_on_server and network != network_on_server
            )
        ):
            # Get the first network that is the same, but has
            # changed locally - that is, the network address and
            # subnet matches but flags, description, or
            # something else is different in the data received
            # from file
            merged = network.merge(
                next(
                    filter(
                        lambda network_on_server: (
                            network in network_on_server
                            and network != network_on_server
                        ),
                        existing_networks,
                    )
                )
            )
            diff["CHANGED_IN_FILE"].append(merged)
    for network in network_list:
        if all(
            (
                network not in network_on_server
                for network_on_server in existing_networks
            )
        ):
            diff["NOT_ON_SERVER"].append(network)
    for network in network_list:
        if any(
            (network == network_on_server for network_on_server in existing_networks)
        ):
            diff["UNMODIFIED"].append(network)
    return diff



import sys
from argus_api.api.cases.v2.case import advanced_case_search, update_case, close_case
from argus_api.api.customers.v2.customer import get_customer
from argus_api.exceptions.http import ArgusException
from argus_cli.helpers.log import log
from argus_cli.plugin import register_command
from argus_cli.utils.time import date_or_relative
from argus_plugins import argus_cli_module
from argus_plugins.cases.utils import (
    get_customer_id,
    CASE_TYPES,
    PRIORITIES,
    STATUSES,
    KEYWORD_FIELDS,
)

TIME_FIELD_STRATEGIES = [
    "createdTimestamp",
    "lastUpdatedTimestamp",
    "closedTimestamp",
    "publishedTimestamp",
    "all",
]


@register_command(extending="cases", module=argus_cli_module)
def change_status(
    start: date_or_relative,
    end: date_or_relative,
    new_status: STATUSES,
    reason: str = None,
    reason_locale_delimiter: str = None,
    keyword: str = None,
    keywords: str = None,
    keyword_field: KEYWORD_FIELDS = None,
    customer: get_customer_id = None,
    exclude_customer: get_customer_id = None,
    type: CASE_TYPES = None,
    service: str = None,
    category: str = None,
    status: STATUSES = None,
    priority: PRIORITIES = None,
    time_field_strategy: TIME_FIELD_STRATEGIES = "lastUpdatedTimestamp",
    dry: bool = False,
):
    """This command can be used to close multiple cases in one go, based on a set of parameters.

    :param start: Time to start filtering the case from (ISO8601 format or relative time)
    :param end: Time to end filtering the case from (ISO8601 format or relative time)
    :param new_status: The new status of the case
    :param reason: The reasoning for the change (If none - it will post "Changing from <old> to <new>")
    :param reason_locale_delimiter: Delimiter for reason locale: left side NOR, right side ENG ("<norwegian text><delimiter><english text>")
    :param list keyword: DEPRECATED, prefer --keywords
    :param list keywords: Keyword(s) to filter by. If using a sentence, remember to encapsulate with quotes.
    :param list keyword_field: The field(s) to filter by
    :param list customer: Customers to include in the search
    :param list exclude_customer: Customers to include in the search
    :param list type: Case type to search for
    :param list service: Case service to search for
    :param list category: Case subcategory type to search for
    :param list status: Status to search for
    :param list priority: Priority to search for
    :param time_field_strategy: The timestamp to search by
    :param dry: Makes the changes NOT commit to the server.
    """
    # combine the "keyword" and "keywords" argument to keep backward compatibility
    # after the arguments has been renamed.
    keywords_combined = None
    if keyword or keywords:
        keywords_combined = keyword or []
        keywords_combined.extend(keywords or [])

    if dry:
        log.info("Dry-run - No changes will be committed")

    if reason_locale_delimiter and (
        not reason or reason_locale_delimiter not in reason
    ):
        print(f'reason "{reason}" missing delimiter "{reason_locale_delimiter}"')
        sys.exit(1)  # Avoid commenting messages with formatting errors to customers

    sub_criteria = []
    if exclude_customer:
        # TODO: This isn't a good solution in the long run. Should create common handling for these kind of arguments.
        sub_criteria.append({"exclude": True, "customerID": exclude_customer})

    log.debug("Fetching cases...")
    cases = advanced_case_search(
        limit=0,
        startTimestamp=start,
        endTimestamp=end,
        timeFieldStrategy=time_field_strategy,
        keywords=keywords_combined,
        keywordFieldStrategy=keyword_field,
        priority=priority,
        status=status,
        type=type,
        category=category,
        service=service,
        customerID=customer,
        subCriteria=sub_criteria,
    )["data"]

    if not cases:
        log.debug(
            f"Advanced cases search returned 0 cases. The status change was {new_status}"
        )
        return

    log.debug("Updating {num_cases} cases.".format(num_cases=len(cases)))
    exit_code = 0
    for case in cases:
        print(
            "Updating case #{case[id]} from {case[status]} to {new_status}".format(
                case=case, new_status=new_status
            )
        )
        if dry:
            continue

        comment = reason
        if comment and reason_locale_delimiter and reason_locale_delimiter in comment:
            customer_language = get_customer(case["customer"]["id"])["data"]["language"]
            comment = reason.split(reason_locale_delimiter)
            comment = comment[0] if customer_language == "norwegian" else comment[1]

        try:
            if new_status == "closed":
                close_case(
                    case["id"],
                    comment=comment
                    or 'Changing status from "{}" to "{}"'.format(
                        case["status"], new_status
                    ),
                )
            else:
                update_case(
                    case["id"],
                    status=new_status,
                    comment=comment
                    or 'Changing status from "{}" to "{}"'.format(
                        case["status"], new_status
                    ),
                )
        except ArgusException as e:
            exit_code = 1
            print(f"Error setting case #{case['id']} status to {new_status}: {e}")
    sys.exit(exit_code)
